@@ -48,6 +48,19 @@ if config.get('logging'):
     logging.basicConfig(level=logging.INFO)
 
 
+@task
+def upload_diff(pr_number, src='/home/erp/src', repository='erp', sudo_user='erp'):
+    temp_dir = '/tmp/%s.diff' % pr_number
+    remote_dir = '{}/{}/patches/'.format(src, repository)
+    sudo("mkdir -p %s" % remote_dir)
+    diff_path = '{}/{}.diff'.format(remote_dir, pr_number)
+    with io.open('deploy/patches/{}.diff'.format(pr_number), 'r', encoding='utf-8') as dfile:
+        logger.info('Uploading diff {}.diff'.format(pr_number))
+        put('deploy/patches/{}.diff'.format(pr_number), temp_dir, use_sudo=True)
+    sudo("mv %s %s" % (temp_dir, diff_path))
+    sudo("chown {0}: {1}".format(sudo_user, diff_path))
+
+
 @task
 def upload_patches(
     pr_number, from_commit=None, src='/home/erp/src', repository='erp', sudo_user='erp'
@@ -80,6 +93,16 @@ def upload_patches(
     sudo("chown -R {0}: {1}".format(sudo_user, remote_dir))
 
 
+@task
+def apply_remote_diff(pr_number, src='/home/erp/src', repository='erp',
+                      sudo_user='erp'
+):
+    with settings(sudo_user=sudo_user):
+        with cd("{}/{}".format(src, repository)):
+            diff_file = 'patches/{}.diff'.format(pr_number)
+            PatchApplier.apply(diff_file)
+
+
 @task
 def apply_remote_patches(
     name, from_patch=0, src='/home/erp/src', repository='erp', sudo_user='erp',
@@ -128,6 +151,43 @@ class WiggleException(Exception):
 class GitHubException(Exception):
     pass
 
+
+class PatchApplier(object):
+
+    @staticmethod
+    def apply(diff, stash=True, reject=False, message=None):
+        need_stash = sudo("test -f .gitignore && git ls-files -om -X .gitignore || git ls-files -om")
+        stashed = False
+        if stash and not need_stash:
+            stash = False
+        if message is None:
+            message = 'Apply {}'.format(diff)
+        if stash:
+            print(colors.yellow('Stashing all before...'))
+            sudo("git stash -u")
+            stashed = True
+        try:
+            if reject:
+                reject = '  --reject'
+            else:
+                reject = ''
+            print(colors.green('Applying diff {}'.format(diff)))
+            sudo(
+                "git apply {}{}".format(diff, reject),
+            )
+            print(colors.green('Commit!'))
+            sudo(
+                'git add -A && git commit -m "{}"'.format(message),
+            )
+        except:
+            print(colors.red('\U000026D4 Error applying diff'))
+            raise
+        finally:
+            if stash and stashed:
+                print(colors.yellow('Unstashing...'))
+                sudo("git stash pop")
+
+
 class GitApplier(object):
     def __init__(self, patches):
         self.patches = patches
@@ -267,9 +327,13 @@ def find_from_to_commits(pr_number, owner='gisce', repository='erp'):
 def export_patches_from_git(from_commit, to_commit, pr_number):
     logger.info('Exporting patches from %s to %s' % (from_commit, to_commit))
     deploy_path = "deploy/patches/{}".format(pr_number)
-    if isdir(deploy_path):
-        local("rm -r {}".format(deploy_path))
-    local("mkdir -p {}".format(deploy_path))
+    try:
+        if isdir(deploy_path):
+            local("rm -r {}".format(deploy_path))
+        local("mkdir -p {}".format(deploy_path))
+    except BaseException as e:
+        logger.error('Permission denied to write {} in the current directory'.format(deploy_path))
+        raise
     local("git format-patch -o deploy/patches/%s %s..%s" % (
         pr_number, from_commit, to_commit)
     )
@@ -308,12 +372,32 @@ def get_commits(pr_number, owner='gisce', repository='erp'):
     return commits
 
 
+@task
+def export_diff_from_github(pr_number, owner='gisce', repository='erp'):
+    diff_path = "deploy/patches/{}.diff".format(pr_number)
+    tqdm.write('Exporting diff from Github')
+    headers = {
+        'Authorization': 'token %s' % github_config()['token'],
+        'Accept': 'application/vnd.github.v3.diff'
+    }
+    url = 'https://api.github.com/repos/{owner}/{repository}/pulls/{pr_number}'.format(
+        owner=owner, repository=repository, pr_number=pr_number
+    )
+    r = requests.get(url, headers=headers)
+    with open(diff_path, 'w') as f:
+        f.write(r.text.encode('utf-8'))
+
+
 @task
 def export_patches_from_github(
     pr_number, from_commit=None, owner='gisce', repository='erp'
 ):
     patch_folder = "deploy/patches/%s" % pr_number
-    local("mkdir -p %s" % patch_folder)
+    try:
+        local("mkdir -p %s" % patch_folder)
+    except BaseException as e:
+        logger.error('Permission denied to write {} in the current directory'.format(patch_folder))
+        raise
     tqdm.write('Exporting patches from GitHub')
     headers = {'Authorization': 'token %s' % github_config()['token']}
     commits = get_commits(pr_number, owner=owner, repository=repository)
@@ -402,7 +486,7 @@ def get_deploys(pr_number, owner='gisce', repository='erp', commit=None):
     pull = json.loads(r.text)
     if commit is None:
         commit = pull['head']['sha']
-    url = "https://api.github.com/repos/{}/{}/deployments?ref={}".format(
+    url = "https://api.github.com/repos/{}/{}/deployments?sha={}".format(
         owner, repository, commit
     )
     r = requests.get(url, headers=headers)
@@ -482,7 +566,11 @@ def mark_deploy_status(
 
 @task
 def export_patches_pr(pr_number, owner='gisce', repository='erp'):
-    local("mkdir -p deploy/patches/%s" % pr_number)
+    try:
+        local("mkdir -p deploy/patches/%s" % pr_number)
+    except BaseException as e:
+        logger.error('Permission denied to write deploy/patches/{} in the current directory'.format(patch_folder))
+        raise
     from_commit, to_commit, branch = find_from_to_commits(
         pr_number, owner=owner, repository=repository
     )
@@ -533,7 +621,8 @@ def check_am_session(src='/home/erp/src', repository='erp', sudo_user='erp'):
 def apply_pr(
         pr_number, from_number=0, from_commit=None, skip_upload=False,
         hostname=False, src='/home/erp/src', owner='gisce', repository='erp',
-        sudo_user='erp', auto_exit=False, force_name=None, re_deploy=False
+        sudo_user='erp', auto_exit=False, force_name=None, re_deploy=False,
+        as_diff=False
 ):
     if force_name:
         repository_name = force_name
@@ -577,30 +666,45 @@ def apply_pr(
             deploy_id
         )))
         if not skip_upload:
-            pass
-            export_patches_from_github(pr_number,
-                                       from_commit,
-                                       owner=owner,
-                                       repository=repository)
-            upload_patches(pr_number,
-                           from_commit,
-                           src=src,
-                           repository=repository_name,
-                           sudo_user=sudo_user)
-        if from_commit:
-            from_ = from_commit
+            if as_diff:
+                export_diff_from_github(
+                    pr_number, owner=owner, repository=repository
+                )
+                upload_diff(
+                    pr_number, src=src, repository=repository,
+                    sudo_user=sudo_user
+                )
+            else:
+                export_patches_from_github(pr_number,
+                                           from_commit,
+                                           owner=owner,
+                                           repository=repository)
+                upload_patches(pr_number,
+                               from_commit,
+                               src=src,
+                               repository=repository_name,
+                               sudo_user=sudo_user)
+        if as_diff:
+            tqdm.write(colors.yellow("Applying diff \U0001F648"))
+            check_am_session(src=src, repository=repository_name)
+            apply_remote_diff(
+                pr_number, src=src, repository=repository, sudo_user=sudo_user
+            )
         else:
-            from_ = from_number
-        tqdm.write(colors.yellow("Applying patches \U0001F648"))
-        check_am_session(src=src, repository=repository_name)
-        result = apply_remote_patches(
-            pr_number,
-            from_,
-            src=src,
-            repository=repository_name,
-            sudo_user=sudo_user,
-            auto_exit=auto_exit,
-        )
+            if from_commit:
+                from_ = from_commit
+            else:
+                from_ = from_number
+            tqdm.write(colors.yellow("Applying patches \U0001F648"))
+            check_am_session(src=src, repository=repository_name)
+            result = apply_remote_patches(
+                pr_number,
+                from_,
+                src=src,
+                repository=repository_name,
+                sudo_user=sudo_user,
+                auto_exit=auto_exit,
+            )
         mark_deploy_status(deploy_id,
                            state='success',
                            owner=owner,
